
#include <Y/y.h>
#include <Y/modules/module.h>
#include <Y/util/yutil.h>
#include <Y/util/index.h>
#include <Y/object/parameters.h>
#include <Y/message/special.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <dlfcn.h>

/* Modules should be changed to a ZOrder, as this will
 * effectively simulate dependencies
 */

struct Module
{
  char *moduleName;
  char *name;
  void *handle;
  void *data;
};

static void moduleUnloadImpl (const char *, struct Module *);

static int
moduleComparisonFunction (struct Module *m1, struct Module *m2)
{
  return strcmp (m1 -> moduleName, m2 -> moduleName);
}

static int
moduleKeyFunction (const char *name, struct Module *m)
{
  return strcmp (name, m -> moduleName);
}

static void
moduleDestructorFunction (struct Module *m)
{
  moduleUnloadImpl (m->moduleName, m);
}

static struct Index *moduleIndex = NULL;

static void moduleSpecial (struct Parameters *);

void
moduleInitialise ()
{
  moduleIndex = indexCreate ((int (*)(void *, void *))moduleKeyFunction,
                             (int (*)(void *, void *))moduleComparisonFunction);
  specialRegister ("module", moduleSpecial);
}

void
moduleFinalise ()
{
  specialUnregister ("module");
  indexDestroy (moduleIndex, (void (*)(void *)) moduleDestructorFunction);
}

void
moduleLoad (const char *moduleName, int argc, const char **argv)
{
  struct Module *module;
  int (*moduleInitialiseFunction)(struct Module *, int, const char **);
  char *moduleFilename;
  int moduleFilenameLength;

  module = indexFind (moduleIndex, (void *)moduleName);
  if (module != NULL)
    {
      Y_ERROR ("Module '%s' already loaded.", moduleName);
      return;
    }

  module = ymalloc (sizeof (struct Module));

  moduleFilenameLength = strlen (yModuleDir) + strlen (moduleName) + 5;
                                                              /* / .so\0 */
  moduleFilename = ymalloc (moduleFilenameLength);
  snprintf (moduleFilename, moduleFilenameLength, "%s/%s.so", yModuleDir,
            moduleName);

  module->moduleName = ystrdup (moduleName);
  module->handle = dlopen (moduleFilename, RTLD_NOW);
  if (module->handle == NULL)
    {
      Y_ERROR ("Module: Failed to load %s: %s\n",
               moduleName, dlerror ());
      abort ();
    }

  yfree (moduleFilename);

  moduleInitialiseFunction = (int (*)(struct Module *, int, const char **))
                             dlsym (module->handle, "initialise");
  if (moduleInitialiseFunction == NULL)
    {
      Y_ERROR ("Module: Failed to find initialise for %s: %s\n",
               moduleName, dlerror ());
      yfree (module);
      return;
    }

  if (moduleInitialiseFunction (module, argc, argv) != 0)
    {
      yfree (module);
      return;
    }
  indexAdd (moduleIndex, module);
}

void
moduleUnload (const char *name)
{
  struct Module *module;
  module = indexRemove (moduleIndex, (void *)name);
  moduleUnloadImpl (name, module);
}

void
moduleUnloadImpl (const char *name, struct Module *module)
{
  int (*moduleFinaliseFunction)(struct Module *);
  Y_TRACE ("Unload Module %s", name);
  if (module == NULL)
    return;
  moduleFinaliseFunction = dlsym(module -> handle, "finalise");
  if (moduleFinaliseFunction != NULL)
    {
      moduleFinaliseFunction (module);
    }
  dlclose (module -> handle);
  yfree (module -> moduleName);
  yfree (module);
}

void
moduleSpecial (struct Parameters *params)
{
  int pCount = parametersGetInputCount (params);

  if (pCount < 1)
    return;

  if (strcasecmp (parametersGetInput (params, 0), "list") == 0)
    {
      struct IndexIterator *iter = indexGetStartIterator (moduleIndex);
      while (indexiteratorHasValue (iter))
        {
          struct Module *module = indexiteratorGet (iter);
          char *buffer = ymalloc (1024);
          snprintf (buffer, 1024, "%s\t%s", module -> moduleName, module -> name);
          parametersAddOutput (params, buffer);
          parametersFreeOnDestroy (params, buffer);
          indexiteratorNext (iter);
        }
      indexiteratorDestroy (iter);
    }

  if (pCount < 2)
    return;

  if (strcasecmp (parametersGetInput (params, 0), "load") == 0)
    {
      const char *argv[pCount - 1];
      int i;
      for (i=0; i < pCount - 2; ++i)
        argv[i] = parametersGetInput (params, i+2);
      argv[pCount-2] = NULL;
      moduleLoad (parametersGetInput (params, 1), pCount-2, argv);
      return;
    }

  if (strcasecmp (parametersGetInput (params, 0), "unload") == 0)
    {
      moduleUnload (parametersGetInput (params, 1));
      return;
    }
}

/* arch-tag: 0644c645-a2c2-4680-b044-3090714ed36c
 */
