#include <Y/screen/screen.h>
#include <Y/util/yutil.h>
#include <Y/util/index.h>
#include <Y/input/arrow.h>
#include <Y/message/special.h>

#include <stdio.h>
#include <string.h>

static struct Index *viewports;
static struct Widget *rootWidget = NULL;
static struct Rectangle *screenRectangle = NULL;

static int
viewportsKeyFunction (void *key_v, void *obj_v)
{
  int key = *((int *)key_v);
  struct Viewport *obj = obj_v;
  int objKey = viewportGetID (obj);
  if (key == objKey)
    return 0;
  if (key < objKey)
    return -1;
  else
    return 1; 
}

static int
viewportsComparisonFunction (void *obj1_v, void *obj2_v)
{
  struct Viewport *obj1 = obj1_v;
  struct Viewport *obj2 = obj2_v;
  int obj1Key = viewportGetID (obj1);
  int obj2Key = viewportGetID (obj2);
  if (obj1Key == obj2Key)
    return 0;
  if (obj1Key < obj2Key)
    return -1;
  else
    return 1; 
}

static void
screenViewportSpecial (struct Parameters *params)
{
  if (parametersGetInputCount (params) > 0)
    {
      const char *type = parametersGetInput(params, 0);
      if (strcmp (type, "list") == 0)
        {
          struct IndexIterator *iter = indexGetStartIterator (viewports);
          while (indexiteratorHasValue (iter))
            {
              struct Viewport *vp = indexiteratorGet (iter);
              const char *vpName = viewportGetName (vp);
              char *buffer = ymalloc (strlen (vpName) + 20);
              sprintf (buffer, "%d\t%s", viewportGetID (vp), vpName);
              parametersAddOutput (params, buffer);
              parametersFreeOnDestroy (params, buffer);
              indexiteratorNext (iter);
            }
          indexiteratorDestroy (iter);
        }
      else
        {
          int vpid = strtoul (type, NULL, 10);
          struct Viewport *vp = indexFind (viewports, &vpid);
          if (vp != NULL)
            {
              parametersShiftInput (params, 1);
              viewportSpecial (vp, params);
            }
        }
    }
}

void
screenInitialise ()
{
  viewports = indexCreate (viewportsKeyFunction, viewportsComparisonFunction);
  screenRectangle = rectangleCreate (0, 0, 800, 600);
  specialRegister ("viewport", screenViewportSpecial);
}

void
screenRegisterViewport  (struct Viewport *v)
{
  indexAdd (viewports, v);
  screenViewportsChanged ();
  screenInvalidateRectangle (viewportGetRectangle (v));
}

void
screenUnregisterViewport(struct Viewport *v)
{
  indexRemove (viewports, v);
  screenViewportsChanged ();
}

void
screenConstrainPoint (int *x_p, int *y_p)
{
  /* constrians the point (x, y) to fall within the viewports */
  /* the point is moved the minimum distance possible */

  int candidateX = *x_p;
  int candidateY = *y_p;
  int distance = -1;
  struct IndexIterator *iter;

  iter = indexGetStartIterator (viewports);
  while (indexiteratorHasValue (iter) && distance != 0)
    {
      struct Rectangle *vRect;
      int nx, ny, nd;
      vRect = viewportGetRectangle (indexiteratorGet (iter));
      if (*x_p < vRect -> x)
        nx = vRect -> x;
      else if (*x_p > vRect -> x + vRect -> w - 1)
        nx = vRect -> x + vRect -> w - 1;
      else
        nx = *x_p;
      if (*y_p < vRect -> y)
        ny = vRect -> y;
      else if (*y_p > vRect -> y + vRect -> h - 1)
        ny = vRect -> y + vRect -> h - 1;
      else
        ny = *y_p;
      nd = (nx - *x_p) * (nx - *x_p)  +  (ny - *y_p) * (ny - *y_p);
      if (distance < 0 || nd < distance)
        {
          candidateX = nx;
          candidateY = ny;
          distance = nd;
        }

      rectangleDestroy (vRect);
      indexiteratorNext (iter);      
    }
  indexiteratorDestroy (iter);

  *x_p = candidateX;
  *y_p = candidateY;

}

void
screenSetRootWidget (struct Widget *w)
{
  rootWidget = w;
  widgetMove (rootWidget, screenRectangle -> x, screenRectangle -> y);
  widgetResize (rootWidget, screenRectangle -> w, screenRectangle -> h);
}

struct Widget *
screenGetRootWidget ()
{
  return rootWidget;
}

void 
screenInvalidateRectangle (struct Rectangle *r)
{
  struct IndexIterator *iterator;

  /* for each viewport, call viewportInvalidateRectangle */
  iterator = indexGetStartIterator (viewports);
  while (indexiteratorHasValue (iterator))
    {
      viewportInvalidateRectangle (indexiteratorGet (iterator), r);
      indexiteratorNext (iterator);
    }
  indexiteratorDestroy (iterator);

  rectangleDestroy (r);
}

void
screenViewportsChanged ()
{
  /* determine the bounding box of all viewports */
  struct IndexIterator *iterator;
  struct Rectangle *newRectangle;

  if (indexCount (viewports) == 0)
    return;

  iterator = indexGetStartIterator (viewports);

  if (!indexiteratorHasValue (iterator))
    return;

  newRectangle = viewportGetRectangle (indexiteratorGet (iterator));

  indexiteratorNext (iterator);
  while (indexiteratorHasValue (iterator))
    {
      struct Rectangle *viewportRectangle;
      viewportRectangle = viewportGetRectangle (indexiteratorGet (iterator));
      rectangleUnion (newRectangle, newRectangle, viewportRectangle);
      rectangleDestroy (viewportRectangle);
      indexiteratorNext (iterator);
    }
  indexiteratorDestroy (iterator);

  rectangleDestroy (screenRectangle);
  screenRectangle = newRectangle;

  widgetMove (rootWidget, screenRectangle -> x, screenRectangle -> y);
  widgetResize (rootWidget, screenRectangle -> w, screenRectangle -> h);

}

void
screenUpdate ()
{
  struct IndexIterator *iterator;

  /* for each viewport, call viewportUpdate */
  iterator = indexGetStartIterator (viewports);
  while (indexiteratorHasValue (iterator))
    {
      viewportUpdate (indexiteratorGet (iterator));
      indexiteratorNext (iterator);
    }
  indexiteratorDestroy (iterator);

}

void
screenRender (struct Renderer *renderer)
{
  /* How to Render:
   * 1. Paint the rootWidget (and it's non-buffered children)
   *    directly onto the renderer.
   * 2. Render the widget tree (i.e. all the buffers) onto the
   *    renderer in ZOrder.
   * 3. Paint/Render the Overlay Widgets onto the Renderer
   *    in ZOrder.
   * 3. Paint the mouse pointer onto the renderer.
   */
#if 0
  struct Painter *painter= rendererGetPainter (renderer, rect);
#endif
  
  if (rootWidget != NULL)
    {
      struct Rectangle *rect = widgetGetRectangle (rootWidget);
      rendererEnter (renderer, rect, rect->x, rect->y);
#if 0
      widgetPaint (rootWidget, painter);
#endif
      widgetRender (rootWidget, renderer);
      rendererLeave (renderer);
      rectangleDestroy (rect);
    }
  else
    {
      rendererDrawFilledRectangle (renderer, 0xFF808080,
                                   screenRectangle->x, screenRectangle->y,
                                   screenRectangle->w, screenRectangle->h);
    }
  {
    int pointerX, pointerY;
    struct Rectangle pointer;
    pointerGetPosition (&pointerX, &pointerY);
    pointer.x = pointerX;
    pointer.y = pointerY;
    pointer.w = 32;
    pointer.h = 32;
    if (rendererEnter (renderer, &pointer, 0, 0))
      {
        rendererBlitRGBAData (renderer, pointer.x, pointer.y,
           ((uint32_t *)pointerArrow.pixel_data) + (pointer.y - pointerY) * 32 + (pointer.x - pointerX),
           pointer.w, pointer.h, 32);
      }
  } 
}

void
screenFinalise ()
{
  rectangleDestroy (screenRectangle);
  indexDestroy (viewports, NULL);
}

/* arch-tag: 7a770377-38cb-4326-8708-a4c8f5033528
 */
