#include <Y/screen/viewport.h>
#include <Y/screen/screen.h>
#include <Y/screen/swrenderer.h>
#include <Y/main/control.h>
#include <Y/util/yutil.h>

#include <stdio.h>
#include <string.h>

struct Viewport
{
  int id;
  struct VideoDriver *video;
  struct List *invalidRectangles;
  int updateEventID;
  int x, y, w, h;
};

static int nextViewportID = 0;

struct Viewport *
viewportCreate (struct VideoDriver *video)
{
  struct Viewport *self = ymalloc (sizeof (struct Viewport));
  self -> id = nextViewportID++;
  self -> video = video;
  self -> invalidRectangles = listCreate ();
  self -> x = 0;
  self -> y = 0;
  video -> getPixelDimensions (video, &(self -> w), &(self -> h)); 
  if (self -> w <= 0)
    self -> w = 800;
  if (self -> h <= 0)
    self -> h = 600;
  self -> updateEventID = 0;
  return self;
}

void
viewportDestroy (struct Viewport *self)
{
  listDestroy (self -> invalidRectangles, (void (*)(void *)) rectangleDestroy);
  yfree (self);
}

int
viewportGetID (struct Viewport *self)
{
  return self -> id;
}

struct Rectangle *
viewportGetRectangle (struct Viewport *self)
{
  struct Rectangle *rect = ymalloc (sizeof (struct Rectangle));
  rect -> x = self -> x;
  rect -> y = self -> y;
  rect -> w = self -> w;
  rect -> h = self -> h;
  return rect;
}

void
viewportInvalidateRectangle (struct Viewport *self, const struct Rectangle *r)
{
  listAppend (self -> invalidRectangles, rectangleDuplicate (r));
  if (self -> updateEventID == 0)
    {
      self -> updateEventID =
        controlTimerDelay (0, 10, self, (void (*)(void *))viewportUpdate);
    }
}

void
viewportSetSize (struct Viewport *self, int w, int h)
{
  struct Rectangle r;
  r.x = self -> x;
  r.y = self -> y;
  r.w = self -> w = w;
  r.h = self -> h = h;
  screenViewportsChanged ();
  viewportInvalidateRectangle (self, &r);
}

void
viewportSpecial (struct Viewport *self, struct Parameters *params)
{
  int pCount = parametersGetInputCount (params);

  if (pCount < 1)
    return;
  if (strcasecmp (parametersGetInput (params, 0), "listResolutions") == 0)
    {
      struct List *resolutions = viewportGetResolutions (self);
      struct ListIterator *iter = listGetStartIterator (resolutions);
      Y_TRACE ("Found %d resolutions", listCount (resolutions));
      while (listiteratorHasValue (iter))
        {
          struct VideoResolution *vres = listiteratorGet (iter);
          char *resName = ystrdup (vres->name);
          parametersAddOutput (params, resName);
          parametersFreeOnDestroy (params, resName);
          listiteratorNext (iter);
        }
      listiteratorDestroy (iter);
      listDestroy (resolutions, NULL);
      return;
    }

  if (pCount < 2)
    return;
  if (strcasecmp (parametersGetInput (params, 0), "call") == 0)
    {
      parametersShiftInput (params, 1);
      if (self -> video -> special != NULL)
        self -> video -> special (self -> video, params);
      return;
    }
  if (strcasecmp (parametersGetInput (params, 0), "setResolution") == 0)
    {
      viewportSetResolution (self, parametersGetInput (params, 1));
      return;
    }

}

const char *
viewportGetName (struct Viewport *self)
{
  if (self -> video -> getName != NULL)
    return self -> video -> getName (self -> video);
  else
    return NULL;
}

struct List *
viewportGetResolutions (struct Viewport *self)
{
  if (self -> video -> getResolutions != NULL)
    return self -> video -> getResolutions (self -> video);
  else
    return NULL;
}

void
viewportSetResolution (struct Viewport *self, const char *name)
{
  if (self -> video -> setResolution != NULL)
    self -> video -> setResolution (self -> video, name);
}

void
viewportUpdate (struct Viewport *self)
{
  struct Rectangle *viewportRectangle = viewportGetRectangle (self);
  struct ListIterator *iter;

  if (listCount (self -> invalidRectangles) == 0)
    return;

  rectanglelistUnionOverlaps (self -> invalidRectangles);

  self -> video -> beginUpdates (self -> video);

  /* for each rectangle in the collection */
  iter = listGetStartIterator (self -> invalidRectangles); 
  while (listiteratorHasValue (iter))
    {
      /* compute the intersected rectangle */
      struct Rectangle *invalid = listiteratorGet (iter);
      struct Rectangle *visible = ymalloc (sizeof (struct Rectangle));
      if (rectangleIntersect (visible, invalid, viewportRectangle))
        {
          /* create a renderer (visitor) and pass it over the widget structure
           */
          struct Renderer *renderer =
                  self -> video -> getRenderer (self -> video, visible);
          screenRender (renderer);
          rendererComplete (renderer);
          rendererDestroy (renderer);
        }
      rectangleDestroy (visible);
      listiteratorNext (iter);
    }

  listiteratorDestroy (iter);
  rectangleDestroy (viewportRectangle);

  self -> video -> endUpdates (self -> video);

  listDestroy (self -> invalidRectangles, (void (*)(void *))rectangleDestroy);
  self -> invalidRectangles = listCreate ();

  self -> updateEventID = 0;

}

/* arch-tag: 24ab628b-2e9c-4bd2-9e84-f5428e71b764
 */
