#include <Y/widget/desktop.h>
#include <Y/widget/widget_p.h>

#include <Y/util/yutil.h>
#include <Y/buffer/rgbabuffer.h>
#include <Y/buffer/painter.h>
#include <Y/screen/screen.h>

#include <Y/object/class_p.h>
#include <Y/object/object_p.h>

#include <Y/modules/windowmanager.h>

#include <Y/text/font.h>

#include <Y/util/zorder.h>

#include <stdio.h>

struct Desktop
{
  struct Widget widget;
  struct ZOrder *windows;
  struct Widget *pointerWidget;
};

static int desktopPointerMotion (struct Widget *, int, int, int, int);
static int desktopPointerButton (struct Widget *, int, int, int, int);
                                
static int desktopKeyboardRaw   (struct Widget *, enum YKeyCode, int, int);
static void desktopRender (struct Widget *, struct Renderer *);
static void desktopResize (struct Widget *);

CLASSDEF struct Class desktopClass =
{
  name:          "Desktop",
  super:         &widgetClass,
  methodCount:   0,
  methods:
    {
    }
};

static struct WidgetTable desktopTable =
{
  c:             &desktopClass,
  pointerMotion: desktopPointerMotion,
  pointerButton: desktopPointerButton,
  keyboardRaw:   desktopKeyboardRaw,
  render:        desktopRender,
  resize:        desktopResize
};

static inline struct Desktop *
castBack (struct Widget *widget)
{
  /* assert ( widget -> c == desktopClass ); */
  return (struct Desktop *)widget;
}

static inline const struct Desktop *
castBackConst (const struct Widget *widget)
{
  /* assert ( widget -> c == desktopClass ); */
  return (const struct Desktop *)widget;
}

static int
windowsKeyFunction (void *key_v, void *obj_v)
{
  int key = *((int *)key_v);
  struct Object *obj = obj_v;
  int objKey = objectGetID (obj);
  if (key == objKey)
    return 0;
  if (key < objKey)
    return -1;
  else
    return 1;
}

static int
windowsComparisonFunction (void *obj1_v, void *obj2_v)
{
  struct Object *obj1 = obj1_v;
  struct Object *obj2 = obj2_v;
  int obj1Key = objectGetID (obj1);
  int obj2Key = objectGetID (obj2);
  if (obj1Key == obj2Key)
    return 0;
  if (obj1Key < obj2Key)
    return -1;
  else
    return 1;
}


struct Desktop *
desktopCreate ()
{
  struct Desktop *self = ymalloc (sizeof (struct Desktop));
  objectInitialise (&(self -> widget.o), &desktopClass);
  widgetInitialise (&(self -> widget), &desktopTable);
  
  self -> widget.x = 0;
  self -> widget.y = 0;
  self -> widget.w = 1;
  self -> widget.h = 1;

  self -> windows =
    zorderCreate (windowsKeyFunction, windowsComparisonFunction);

  self -> pointerWidget = NULL;

  return self;
}

void
desktopDestroy (struct Desktop *self)
{
  zorderDestroy (self -> windows, NULL);
  objectFinalise (desktopToObject (self));
  yfree (self);
}

struct Widget *
desktopToWidget (struct Desktop *self)
{
  return &(self -> widget);
}

struct Object *
desktopToObject (struct Desktop *self)
{
  return &(self -> widget.o);
}

int
desktopPointerMotion (struct Widget *self_w, int x, int y, int dx, int dy)
{
  struct Desktop *self = castBack (self_w);
  struct ZOrderIterator *iterator; 
  iterator = zorderGetTopIterator (self -> windows);
  while (zorderiteratorHasValue (iterator)) 
    {
      struct Widget *widget = zorderiteratorGet (iterator);
      int wx, wy;
      widgetGetPosition (widget, &wx, &wy);
      if (widgetContainsPoint (widget, x - wx, y - wy))
        {
          if (widget != self -> pointerWidget)
            {
              if (self -> pointerWidget != NULL)
                widgetPointerLeave (self -> pointerWidget);
              widgetPointerEnter (widget, x - wx, y - wy);
              self -> pointerWidget =  widget;
            }
          if (widgetPointerMotion (widget, x - wx, y - wy, dx, dy))
            {
              zorderiteratorDestroy (iterator);
              return 1;
            }
        }
      zorderiteratorMoveDown (iterator);
    }
  zorderiteratorDestroy (iterator);
 
  if (self -> pointerWidget != NULL)
    {
      widgetPointerLeave (self -> pointerWidget);
      self -> pointerWidget = NULL;
    } 

  return 1;
}

int
desktopPointerButton (struct Widget *self_w, int x, int y, int b, int pressed)
{
  struct Desktop *self = castBack (self_w);
  struct ZOrderIterator *iterator; 
  iterator = zorderGetTopIterator (self -> windows);
  while (zorderiteratorHasValue (iterator)) 
    {
      struct Widget *widget = zorderiteratorGet (iterator);
      int wx, wy;
      widgetGetPosition (widget, &wx, &wy);
      if (widgetContainsPoint (widget, x - wx, y - wy))
        {
          if (widgetPointerButton (widget, x - wx, y - wy, b, pressed))
            {
              zorderiteratorDestroy (iterator);
              return 1;
            }
        }
      zorderiteratorMoveDown (iterator);
    }
  zorderiteratorDestroy (iterator);
  

  return 1;
}

int
desktopKeyboardRaw (struct Widget *self_w, enum YKeyCode code,
                    int pressed, int modifierState)
{
  /* struct Desktop *self = castBack (self_w); */
  return 1;
}

void
desktopRender (struct Widget *self_w, struct Renderer *renderer)
{
  struct Desktop *self = castBack (self_w);
  struct ZOrderIterator *iter;

  /* This should be in paint?
   */
  rendererDrawFilledRectangle (renderer, 0xFF404080,
                               self -> widget.x, self -> widget.y,
                               self -> widget.w, self -> widget.h);

  /* render the windows */
  iter = zorderGetBottomIterator (self -> windows);

  while (zorderiteratorHasValue (iter))
    {
      struct Widget *widget = zorderiteratorGet (iter);
      struct Rectangle *widgetRectangle = widgetGetRectangle (widget);
      if (rendererEnter (renderer, widgetRectangle,
                         widgetRectangle->x, widgetRectangle->y))
        {
          widgetRender (widget, renderer);
          rendererLeave (renderer);
        }
      rectangleDestroy (widgetRectangle);
      zorderiteratorMoveUp (iter);
    }
  zorderiteratorDestroy (iter);

}

void
desktopResize (struct Widget *self_w)
{
  struct Desktop *self = castBack (self_w);
  struct ZOrderIterator *iter;

  iter = zorderGetBottomIterator (self -> windows);
  while (zorderiteratorHasValue (iter))
    {
      struct Window *win = zorderiteratorGet (iter);
      enum WindowSizeState winState = windowGetSizeState (win);
      struct Rectangle *winRect = widgetGetRectangle (windowToWidget (win));
      struct Rectangle *newRect = rectangleDuplicate (winRect);
      switch (winState)
        {
          case WINDOW_SIZE_NORMAL:
            /* push any off bottom/right windows back on-to the screen,
             * if possible */
            if (newRect -> w > self -> widget.w)
              newRect -> w = self -> widget.w;
            if (newRect -> x + newRect -> w > self -> widget.w)
              newRect -> x = self -> widget.w - newRect -> w;
            if (newRect -> h > self -> widget.h)
              newRect -> h = self -> widget.h;
            if (newRect -> y + newRect -> h > self -> widget.h)
              newRect -> y = self -> widget.h - newRect -> h;
            if (winRect -> x != newRect -> x || winRect -> y != newRect -> y)
              widgetMove (windowToWidget (win), newRect -> x, newRect -> y);
            if (winRect -> w != newRect -> w || winRect -> h != newRect -> h)
              widgetResize (windowToWidget (win), newRect -> w, newRect -> h);
            break;
          case WINDOW_SIZE_MAXIMISE:
            /* just resize it */
            widgetResize (windowToWidget (win), self -> widget.w, self -> widget.h);
            break;
        }
      rectangleDestroy (winRect);
      rectangleDestroy (newRect);
      zorderiteratorMoveUp (iter);
    }
  zorderiteratorDestroy (iter);
}

void
desktopAddWindow (struct Desktop *self, struct Window *win)
{
  zorderAddAtTop (self -> windows, win);
  widgetSetContainer (windowToWidget (win), desktopToWidget (self));
  widgetRepaint (desktopToWidget (self),
                    widgetGetRectangle (windowToWidget (win)));
}

void
desktopRaiseWindow (struct Desktop *self, struct Window *win)
{
  int id = objectGetID (windowToObject (win));
  struct ZOrderIterator *iter = zorderGetTopIterator (self -> windows);
  struct Object *top = zorderiteratorGet (iter);
  if (top != NULL && objectGetID (top) != id)
    {
      zorderMoveToTop (self -> windows, &id);
      widgetRepaint (desktopToWidget (self),
                        widgetGetRectangle (windowToWidget (win)));
    }
  zorderiteratorDestroy (iter);
}

void
desktopRemoveWindow (struct Desktop *self, struct Window *win)
{
  int id = objectGetID (windowToObject (win));
  zorderRemove (self -> windows, &id);
  if (self -> pointerWidget == windowToWidget (win))
    self -> pointerWidget = NULL;
  widgetRerender (desktopToWidget (self),
                  widgetGetRectangle (windowToWidget (win)));
}

void
desktopCycleWindows (struct Desktop *self, int direction)
{
  struct Window *win;
  if (direction == 1)
    {
      win = zorderGetTop (self -> windows);
      if (win == NULL)
        return;
      zorderMoveToBottom (self -> windows, win);
      widgetRerender (windowToWidget (win), NULL);
    }
  else
    {
      win = zorderGetBottom (self -> windows);
      if (win == NULL)
        return;
      zorderMoveToTop (self -> windows, win);
      widgetRerender (windowToWidget (win), NULL);
    }
  win = zorderGetTop (self -> windows);
  if (win != NULL)
    wmSelectWindow (win);
  widgetRerender (windowToWidget (win), NULL);
}

/* arch-tag: c944bbcb-243c-491e-af68-1953384e05ac
 */
