#include <Y/widget/menu.h>
#include <Y/widget/widget_p.h>

#include <Y/util/yutil.h>
#include <Y/util/index.h>
#include <Y/buffer/painter.h>

#include <Y/object/class_p.h>
#include <Y/object/object_p.h>
#include <Y/modules/theme.h>

#include <Y/text/font.h>

#include <stdio.h>
#include <ctype.h>
#include <string.h>

/*
 * MenuItems are always sorted in id order. There may be gaps in the id
 * sequence.
 * 
 * Text of "-" means a separator.
 * 
 * submenu of NULL means a leaf item that will emit a "clicked" signal with
 * parameter 'id'.
 */

struct MenuItem
{
  int id;
  char *text;
  int yPosition;
  int height;
  struct Menu *submenu;
};

struct Menu
{
  struct Widget widget;
  struct Index *items;  /* index of struct MenuItem * */
  struct MenuItem *current;
};

static int
menuitemKeyFunction (int *key, struct MenuItem *item)
{
  if (*key < item->id)
    return -1;
  else
    return (*key > item->id);
}

static int
menuitemComparisonFunction (struct MenuItem *item1, struct MenuItem *item2)
{
  if (item1->id < item2->id)
    return -1;
  else
    return (item1->id > item2->id);
}

static void
menuitemDestructorFunction (struct MenuItem *item)
{
  yfree (item -> text);
  yfree (item);
}

static void menuReconfigure (struct Widget *);
static void menuPaint (struct Widget *, struct Painter *);
static int menuPointerMotion (struct Widget *, int, int, int, int);
static int menuPointerButton (struct Widget *, int, int, int, int);
static void menuPointerEnter (struct Widget *, int, int);
static void menuPointerLeave (struct Widget *);

static struct Object *menuInstantiate (struct Parameters *);

static void menuCAddItem (struct Menu *, struct Parameters *);

CLASSDEF struct Class menuClass =
{
  name:          "Menu",
  super:         &widgetClass,
  instantiate:   menuInstantiate,
  destroy:       (void (*)(struct Object *))menuDestroy,
  methodCount:   1,
  methods:
    {
      { "addItem", (ObjectMethod_t) menuCAddItem }
    }
};

static struct WidgetTable menuTable =
{
  c:             &menuClass,
  reconfigure:   menuReconfigure,
  paint:         menuPaint,
  pointerMotion: menuPointerMotion,
  pointerButton: menuPointerButton,
  pointerEnter:  menuPointerEnter,
  pointerLeave:  menuPointerLeave
};

static inline struct Menu *
castBack (struct Widget *widget)
{
  /* assert ( widget -> c == menuClass ); */
  return (struct Menu *)widget;
}

static inline const struct Menu *
castBackConst (const struct Widget *widget)
{
  /* assert ( widget -> c == menuClass ); */
  return (const struct Menu *)widget;
}

static void
menuPaint (struct Widget *self_w, struct Painter *painter)
{
  struct Menu *self = castBack (self_w);
  struct Font *font;
  int yp;
  int ascender, offset;
  struct IndexIterator *iter;

  font = themeGetDefaultFont ();
  fontGetMetrics (font, &ascender, NULL, NULL);
  iter = indexGetStartIterator (self -> items);
  while (indexiteratorHasValue (iter))
    {
      struct MenuItem *item = indexiteratorGet (iter);
      yp = item -> yPosition + 2;
      if (strcmp (item -> text, "-") == 0)
        {
          painterSetPenColour (painter, 0x80000000);
          painterDrawHLine (painter, 0, yp - 1, self -> widget.w);
          painterSetPenColour (painter, 0x80FFFFFF);
          painterDrawHLine (painter, 0, yp, self -> widget.w);
        }
      else
        {
          painterSaveState (painter);
          if (self -> current == item)
            themeDrawButtonPane (painter, 0, item -> yPosition,
                                 self -> widget.w, item -> height,
                                 self -> widget.state);
          painterSetPenColour (painter, 0xFF000000);
          fontMeasureString (font, item -> text, &offset, NULL, NULL);
          fontRenderString (font, painter, item -> text, 4 - offset, yp + ascender);
          painterRestoreState (painter);
        }
      indexiteratorNext (iter);
    }
  indexiteratorDestroy (iter);
}

struct Menu *
menuCreate ()
{
  struct Menu *self = ymalloc (sizeof (struct Menu));
  objectInitialise (&(self -> widget.o), &menuClass);
  widgetInitialise (&(self -> widget), &menuTable);
  self -> items = indexCreate ((int (*)(void *, void *)) menuitemKeyFunction,
                        (int (*)(void *, void *)) menuitemComparisonFunction);
  self -> current = NULL;
  return self;
}

void
menuDestroy (struct Menu *self)
{
  indexDestroy (self -> items, (void (*)(void *)) menuitemDestructorFunction);
  widgetFinalise (menuToWidget (self));
  objectFinalise (menuToObject (self));
  yfree (self);
}

struct Widget *
menuToWidget (struct Menu *self)
{
  return &(self -> widget);
}

struct Object *
menuToObject (struct Menu *self)
{
  return &(self -> widget.o);
}

static struct Object *
menuInstantiate (struct Parameters *p)
{
  return menuToObject (menuCreate ());
}

static void menuReconfigure (struct Widget *self_w)
{
  struct Menu *self = castBack (self_w);
  struct Font *font;
  int maxWidth = 0;
  int height = 0;
  int ascender, descender, linegap, width, offset;
  struct IndexIterator *iter;
  font = themeGetDefaultFont ();
  fontGetMetrics (font, &ascender, &descender, &linegap);
  iter = indexGetStartIterator (self -> items);
  while (indexiteratorHasValue (iter))
    {
      struct MenuItem *item = indexiteratorGet (iter);
      if (strcmp (item -> text, "-") == 0)
        {
          item -> yPosition = height;
          item -> height = 4;
          height += 4;
        }
      else
        {
          fontMeasureString (font, item -> text, &offset, &width, NULL);
          if (width - offset > maxWidth)
            maxWidth = width - offset;
          item -> yPosition = height;
          item -> height = ascender + descender + 4;
          height += ascender + descender + 4;
        }
      indexiteratorNext (iter);
    }
  indexiteratorDestroy (iter);

  self -> widget.minWidth  = maxWidth + 8;
  self -> widget.minHeight = height;
  self -> widget.reqWidth  = maxWidth + 8;
  self -> widget.reqHeight = height;
  self -> widget.maxHeight = height;

  widgetReconfigure (self -> widget.container);
}

static struct MenuItem *
menuItemAt (struct Menu *self, int x, int y)
{
  struct IndexIterator *iter = indexGetStartIterator (self -> items);
  while (indexiteratorHasValue (iter))
    {
      struct MenuItem *item = indexiteratorGet (iter);
      if (item -> yPosition <= y && y < item -> yPosition + item -> height)
        {
          indexiteratorDestroy (iter);
          return item;
        }

      indexiteratorNext (iter);
    }
  indexiteratorDestroy (iter);
  return NULL; 
}

int
menuPointerButton (struct Widget *self_w, int x, int y, int b, int pressed)
{
  struct Menu *self = castBack (self_w);
  if (b==0)
    {
      if (pressed)
        {
          pointerGrab (menuToWidget (self));
          self -> widget.state = WIDGET_STATE_PRESSED;
          self -> current = menuItemAt (self, x, y);
          if (self -> current != NULL)
            widgetRepaint (menuToWidget (self),
                rectangleCreate (0, self -> current -> yPosition,
                    self -> widget.w, self -> current -> height));
        }
      else
        {
          pointerRelease ();
          if (self -> current != NULL)
            {
              struct Parameters *params = parametersCreate ();
              parametersAddOutputInteger (params, self -> current -> id);
              objectEmitSignal (menuToObject (self), "clicked", params);
              parametersDestroy (params);
              widgetRepaint (menuToWidget (self),
                  rectangleCreate (0, self -> current -> yPosition,
                      self -> widget.w, self -> current -> height));
            }
          if (widgetContainsPoint (menuToWidget (self), x, y))
            self -> widget.state = WIDGET_STATE_HOVER;
          else
            {
              self -> widget.state = WIDGET_STATE_NORMAL;
              self -> current = NULL;
            }
        }
    }
  return 1; 
}

int
menuPointerMotion (struct Widget *self_w, int x, int y, int dx, int dy)
{
  struct Menu *self = castBack (self_w);
  if (self -> widget.state == WIDGET_STATE_PRESSED &&
      !widgetContainsPoint (menuToWidget (self), x, y))
    {
      self -> widget.state = WIDGET_STATE_CANCELLING;
      if (self -> current != NULL)
        widgetRepaint (menuToWidget (self),
            rectangleCreate (0, self -> current -> yPosition,
                self -> widget.w, self -> current -> height));
      self -> current = NULL;
    }
  else if (widgetContainsPoint (menuToWidget (self), x, y))
    {
      struct MenuItem *item = menuItemAt (self, x, y);
      if (item != NULL && strcmp (item -> text, "-") == 0)
        item = NULL;

      if (item != NULL && self -> widget.state == WIDGET_STATE_CANCELLING)
        self -> widget.state = WIDGET_STATE_PRESSED;

      if (item != self -> current)
        {
          if (self -> current != NULL)
            {
              widgetRepaint (menuToWidget (self),
                  rectangleCreate (0, self -> current -> yPosition,
                      self -> widget.w, self -> current -> height));
            }
          if (item != NULL)
            {
              widgetRepaint (menuToWidget (self),
                  rectangleCreate (0, item -> yPosition,
                      self -> widget.w, item -> height));
            }
          self -> current = item;
        }
      
    }
  return 1;
}

void
menuPointerEnter (struct Widget *self_w, int x, int y)
{
  struct Menu *self = castBack (self_w);
  self -> widget.state = WIDGET_STATE_HOVER;
  self -> current = menuItemAt (self, x, y);
  if (self -> current != NULL)
    widgetRepaint (menuToWidget (self),
        rectangleCreate (0, self -> current -> yPosition,
            self -> widget.w, self -> current -> height));
}

void
menuPointerLeave (struct Widget *self_w)
{
  struct Menu *self = castBack (self_w);
  self -> widget.state = WIDGET_STATE_NORMAL;
  if (self -> current != NULL)
    widgetRepaint (menuToWidget (self),
        rectangleCreate (0, self -> current -> yPosition,
            self -> widget.w, self -> current -> height));
  self -> current = NULL; 
}  

void
menuCAddItem (struct Menu *self, struct Parameters *params)
{
  /* parameters are:  id   text   submenu
   */
  if (parametersGetInputCount (params) == 3)
    {
      struct MenuItem *item;
      int id, submenuId;
      const char *text;
      struct Object *submenu;
      id        = strtol (parametersGetInput (params, 0), NULL, 0);
      text      =         parametersGetInput (params, 1);
      submenuId = strtol (parametersGetInput (params, 2), NULL, 0);

      if (indexFind (self -> items, &id) != NULL)
        return;

      submenu = objectFind (submenuId);
      if (submenu != NULL && !classInherits (objectClass (submenu), "Menu"))
        return; 

      item = ymalloc (sizeof (struct MenuItem));
      item -> id = id;
      item -> text = ystrdup (text);
      item -> height = 0;
      item -> submenu = (struct Menu *) submenu;  /* ok, because we checked */ 

      indexAdd (self -> items, item);
      widgetReconfigure (menuToWidget (self));
    }
}

/* arch-tag: 550b1df2-ba00-4c70-942e-2591256f5eff
 */
