
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <Y/modules/videodriver_interface.h>
#include <Y/modules/module_interface.h>
#include <Y/main/control.h>
#include <Y/screen/viewport.h>
#include <Y/screen/screen.h>
#include <Y/screen/swrenderer.h>
#include <Y/screen/simplerenderer.h>
#include <Y/input/pointer.h>
#include <Y/util/colour.h>
#include <Y/util/yutil.h>
#include <SDL.h>

#define SDL_EVENT_POLL_FREQUENCY 10

struct SDLVideoDriverData
{
  SDL_Surface *sdlSurface;
  struct Viewport *viewport;
  int pollingID;
  int renderSimply;
  struct VideoResolution curRes;
};

static inline struct SDLVideoDriverData *
sdlData (struct VideoDriver *self)
{
  return (struct SDLVideoDriverData *)(self -> d);
}

static void
sdlEventProbe (void *data_v)
{
  int resized = 0, resizedWidth = 0, resizedHeight = 0;
  struct SDLVideoDriverData *data = data_v;
  SDL_Event event;
  if (SDL_PollEvent (NULL))
    {
      while (SDL_PollEvent (&event))
        switch (event.type)
          {
            case SDL_MOUSEMOTION:
              {
                pointerSetPosition (event.motion.x, event.motion.y);
              }
            break;
            case SDL_MOUSEBUTTONDOWN:
            case SDL_MOUSEBUTTONUP:
              {
                int button = 0;
                int state = 0;
                switch (event.button.button)
                  {
                    case SDL_BUTTON_LEFT:   button = 0; break;
                    case SDL_BUTTON_MIDDLE: button = 1; break;
                    case SDL_BUTTON_RIGHT:  button = 2; break;
                  }
                if (event.button.state == SDL_PRESSED)
                  state = 1;
                else
                  state = 0;
                pointerButtonChange (button, state);
              }
            break;

            case SDL_KEYDOWN:
              keyboardKeyDown (event.key.keysym.sym);
              break;
            case SDL_KEYUP:
              keyboardKeyUp (event.key.keysym.sym);
              break;

            case SDL_VIDEORESIZE:
              resized = 1;
              resizedWidth = event.resize.w;
              resizedHeight = event.resize.h;
              break;

            case SDL_QUIT:
              controlShutdownY ();
              break;
          }
    }
 
  if (resized)
    {
      data -> sdlSurface =
        SDL_SetVideoMode (resizedWidth, resizedHeight, 32, SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_RESIZABLE);
      viewportSetSize (data -> viewport, resizedWidth, resizedHeight);
    }
  data -> pollingID = controlTimerDelay (0, SDL_EVENT_POLL_FREQUENCY, data_v, sdlEventProbe);
}


static void
sdlGetPixelDimensions (struct VideoDriver *self, int *x, int *y)
{
  *x = sdlData(self) -> sdlSurface -> w;
  *y = sdlData(self) -> sdlSurface -> h;
}

static const char *
sdlGetName (struct VideoDriver *self)
{
  return self -> module -> name;
}

static struct List *
sdlGetResolutions (struct VideoDriver *self)
{
  struct SDLVideoDriverData *data = sdlData (self);
  struct List *resolutions = listCreate ();
  yfree (data -> curRes.name);
  data -> curRes.name = ymalloc (50);
  sprintf (data -> curRes.name, "%dx%d",
           data -> sdlSurface -> w, data -> sdlSurface -> h);
  listAppend (resolutions, &(data -> curRes));
  return resolutions;
}

static void
sdlSpecial (struct VideoDriver *self, struct Parameters *params)
{
  struct SDLVideoDriverData *data = sdlData (self);
  int pCount = parametersGetInputCount (params);

  if (pCount < 1)
    return;

  if (strcasecmp (parametersGetInput (params, 0), "simpleRender") == 0)
    {
      struct Rectangle *r = viewportGetRectangle (data -> viewport);
      data -> renderSimply = 1;
      viewportInvalidateRectangle (data -> viewport, r);
      rectangleDestroy (r);
      return;
    }
  if (strcasecmp (parametersGetInput (params, 0), "fullRender") == 0)
    {
      struct Rectangle *r = viewportGetRectangle (data -> viewport);
      data -> renderSimply = 0;
      viewportInvalidateRectangle (data -> viewport, r);
      rectangleDestroy (r);
      return;
    } 
}

static void
sdlBeginUpdates (struct VideoDriver *self)
{
}

static void
sdlEndUpdates (struct VideoDriver *self)
{
  SDL_Flip (sdlData(self) -> sdlSurface);
}

static void
sdlDrawPixel (struct VideoDriver *self, uint32_t col, int x, int y)
{
  unsigned pixel;
  Uint32 *bufp;

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    if ( SDL_LockSurface(sdlData(self) -> sdlSurface) < 0 )
      return;
  bufp = (Uint32 *)sdlData(self) -> sdlSurface->pixels + y*sdlData(self) -> sdlSurface->pitch/4 + x;

  colourBlendSourceOver (&pixel, *bufp << 8, col, 0xFF);
  *bufp = pixel;

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    SDL_UnlockSurface(sdlData(self) -> sdlSurface);

}

static void
sdlDrawRectangle (struct VideoDriver *self,
                       uint32_t colour, int x1, int y1, int x2, int y2)
{
  int i;
  Uint32 *bufp1;
  Uint32 *bufp2;

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    if ( SDL_LockSurface(sdlData(self) -> sdlSurface) < 0 )
      return;

  bufp1 = (Uint32 *)sdlData(self) -> sdlSurface->pixels + (y1)*sdlData(self) -> sdlSurface->pitch/4 + x1;
  bufp2 = (Uint32 *)sdlData(self) -> sdlSurface->pixels + (y2)*sdlData(self) -> sdlSurface->pitch/4 + x1;
  for (i = x1; i < x2; ++i)
    {
      *bufp1 = colour;
      *bufp2 = colour;
      bufp1 ++;
      bufp2 ++;
    }
  bufp1 = (Uint32 *)sdlData(self) -> sdlSurface->pixels + (y1)*sdlData(self) -> sdlSurface->pitch/4 + x1;
  bufp2 = (Uint32 *)sdlData(self) -> sdlSurface->pixels + (y2)*sdlData(self) -> sdlSurface->pitch/4 + x1;
  for (i = y1; i < y2; ++i)
    {
      *bufp1 = colour;
      *bufp2 = colour;
      bufp1 += sdlData(self) -> sdlSurface->pitch/4;
      bufp2 += sdlData(self) -> sdlSurface->pitch/4;
    }

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    SDL_UnlockSurface(sdlData(self) -> sdlSurface);
}

static void
sdlDrawFilledRectangle (struct VideoDriver *self, uint32_t colour,
                        int x1, int y1, int x2, int y2)
{

  Uint32 *bufp;
  int i, j;

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    if ( SDL_LockSurface(sdlData(self) -> sdlSurface) < 0 )
      return;

  for (j = y1; j < y2; ++j)
    {
      bufp = (Uint32 *)sdlData(self) -> sdlSurface->pixels + (j)*sdlData(self) -> sdlSurface->pitch/4 + x1;
      for (i = x1; i < x2; ++i)
        {
          *bufp = colour;
          ++bufp;
        }
    }

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    SDL_UnlockSurface(sdlData(self) -> sdlSurface);

}

static void
sdlBlit (struct VideoDriver *self, uint32_t *data,
         int x, int y, int w, int h, int stepping)
{

  uint32_t * bufferLine;
  Uint32 *bufp;
  int i, j;

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    if ( SDL_LockSurface(sdlData(self) -> sdlSurface) < 0 )
      return;

  for (j = 0; j < h; ++j)
    {
      bufferLine = data + (j * stepping);
      bufp = (Uint32 *)sdlData(self) -> sdlSurface->pixels + (y+j)*sdlData(self) -> sdlSurface->pitch/4 + x;
      for (i = 0; i < w; ++i)
        {
          *bufp = *bufferLine; /* colourBlend (*bufp << 8, *bufferLine, *bufferLine & 0xFF); */
          ++bufferLine;
          ++bufp;
        }
    }

  if ( SDL_MUSTLOCK(sdlData(self) -> sdlSurface) )
    SDL_UnlockSurface(sdlData(self) -> sdlSurface);

}


static struct Renderer *
sdlGetRenderer (struct VideoDriver *self, const struct Rectangle *rect)
{
  if (sdlData (self) -> renderSimply)
    return simplerendererGetRenderer (simplerendererCreate (self, rect));
  else
    return swrendererGetRenderer (swrendererCreate (self, rect));
}

int
initialise (struct Module *module, int argc, const char **argv)
{
  struct VideoDriver *videodriver;

  if ( SDL_Init(SDL_INIT_EVERYTHING) < 0 )
    {
        Y_ERROR ("sdl: Unable to init: %s", SDL_GetError());
        return 1;
    }

  videodriver = ymalloc (sizeof (struct VideoDriver));
  videodriver -> d = ymalloc (sizeof (struct SDLVideoDriverData)); 
  videodriver -> module = module;

  sdlData(videodriver) -> sdlSurface = SDL_SetVideoMode(800, 600, 32, SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_RESIZABLE);

  SDL_WM_SetCaption ("Y on SDL", NULL);

  videodriver -> getPixelDimensions = sdlGetPixelDimensions;
  videodriver -> getName = sdlGetName;
  videodriver -> getResolutions = sdlGetResolutions;
  videodriver -> special = sdlSpecial;
  videodriver -> beginUpdates = sdlBeginUpdates;
  videodriver -> endUpdates = sdlEndUpdates;
  videodriver -> drawPixel = sdlDrawPixel;
  videodriver -> drawRectangle = sdlDrawRectangle;
  videodriver -> drawFilledRectangle = sdlDrawFilledRectangle;
  videodriver -> blit = sdlBlit;  
  videodriver -> getRenderer = sdlGetRenderer;

  module -> name = "SDL Video Driver";
  module -> data = videodriver;

  sdlData (videodriver) -> curRes.name = NULL;
  sdlData (videodriver) -> viewport = viewportCreate (videodriver);
  sdlData (videodriver) -> renderSimply = 0;
  screenRegisterViewport (sdlData (videodriver) -> viewport);

  SDL_ShowCursor (SDL_DISABLE);
 
  sdlData (videodriver) -> pollingID =
    controlTimerDelay (0, SDL_EVENT_POLL_FREQUENCY, sdlData(videodriver), sdlEventProbe);

  return 0;
}

int
finalise (struct Module *module)
{
  struct VideoDriver *videodriver = module -> data;
  keyboardReset ();
  screenUnregisterViewport (sdlData (videodriver) -> viewport);
  viewportDestroy (sdlData (videodriver) -> viewport);
  controlCancelTimerDelay (sdlData (videodriver) -> pollingID);
  yfree(sdlData(videodriver));
  yfree(videodriver);
  SDL_ShowCursor (SDL_ENABLE);
  SDL_Quit ();
  return 0;
}

/* arch-tag: 66bb3c53-a3a8-4fe7-806f-d653d440e267
 */
